#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import scipy.sparse
import PIL.Image
import pyamg


def blend(img_target, img_source, img_mask, offset=(0, 0)):
    # compute regions to be blended
    region_source = (
        max(-offset[0], 0),
        max(-offset[1], 0),
        min(img_target.shape[0] - offset[0], img_source.shape[0]),
        min(img_target.shape[1] - offset[1], img_source.shape[1]))
    region_target = (
        max(offset[0], 0),
        max(offset[1], 0),
        min(img_target.shape[0], img_source.shape[0] + offset[0]),
        min(img_target.shape[1], img_source.shape[1] + offset[1]))
    region_size = (region_source[2] - region_source[0], region_source[3] - region_source[1])

    # clip and normalize mask image
    img_mask = img_mask[region_source[0]:region_source[2], region_source[1]:region_source[3]]
    img_mask[img_mask == 0] = False
    img_mask[img_mask != False] = True

    # create coefficient matrix
    A = scipy.sparse.identity(np.prod(region_size), format='lil')
    for y in range(region_size[0]):
        for x in range(region_size[1]):
            if img_mask[y, x]:
                index = x + y * region_size[1]
                A[index, index] = 4
                if index + 1 < np.prod(region_size):
                    A[index, index + 1] = -1
                if index - 1 >= 0:
                    A[index, index - 1] = -1
                if index + region_size[1] < np.prod(region_size):
                    A[index, index + region_size[1]] = -1
                if index - region_size[1] >= 0:
                    A[index, index - region_size[1]] = -1
    A = A.tocsr()

    # create poisson matrix for b
    P = pyamg.gallery.poisson(img_mask.shape)

    # for each layer (ex. RGB)
    for num_layer in range(img_target.shape[2]):
        # get subimages
        t = img_target[region_target[0]:region_target[2], region_target[1]:region_target[3], num_layer]
        s = img_source[region_source[0]:region_source[2], region_source[1]:region_source[3], num_layer]
        t = t.flatten()
        s = s.flatten()

        # create b
        b = P * s
        for y in range(region_size[0]):
            for x in range(region_size[1]):
                if not img_mask[y, x]:
                    index = x + y * region_size[1]
                    b[index] = t[index]

        # solve Ax = b
        x = pyamg.solve(A, b, verb=False, tol=1e-10)

        # assign x to target image
        x = np.reshape(x, region_size)
        x[x > 255] = 255
        x[x < 0] = 0
        x = np.array(x, img_target.dtype)
        img_target[region_target[0]:region_target[2], region_target[1]:region_target[3], num_layer] = x

    return img_target


def test():
    img_mask = np.asarray(PIL.Image.open('./testimages/test1_mask.png'))
    img_mask.flags.writeable = True
    img_source = np.asarray(PIL.Image.open('./testimages/test1_src.png'))
    img_source.flags.writeable = True
    img_target = np.asarray(PIL.Image.open('./testimages/test1_target.png'))
    img_target.flags.writeable = True
    img_ret = blend(img_target, img_source, img_mask, offset=(40, -30))
    img_ret = PIL.Image.fromarray(np.uint8(img_ret))
    img_ret.save('./testimages/test1_ret.png')


if __name__ == '__main__':
    test()



import logging

import Tkinter
import tkFileDialog
import tkMessageBox

import PIL.Image
import PIL.ImageTk
import numpy as np

from pypoi import poissonblending
from pypoi.image_managers import SourceImageManager, DestinationImageManager
from pypoi.util import resource_path as rp


SAVE_MASK_ENABLED = False  # Show 'Save mask image' button

logger = logging.getLogger('GUI')


class PoissonBlendingApp(Tkinter.Tk):
    NUM_OF_EXAMPLES = 4  # Examples in the testimages folder.

    def __init__(self, parent):
        Tkinter.Tk.__init__(self, parent)
        self.src_img_manager = SourceImageManager()
        self.dst_img_manager = DestinationImageManager(self.src_img_manager)

        self.parent = parent
        self.entry = None
        self.label_text = None
        self.entry_text = None

        self.image_result = None

        self.initialize()

    def initialize(self):
        self.create_menu()
        self.create_widgets()

        self.src_img_manager.set_path(rp('./testimages/test1_src.png'))
        self.src_img_manager.load(rp('./testimages/test1_mask.png'))

        self.dst_img_manager.set_path(rp('./testimages/test1_target.png'))
        self.dst_img_manager.load()
        from pypoi.testimages.config import offset
        self.dst_img_manager.offset = offset[0]

        self.src_img_manager.draw()

    def create_widgets(self):
        self.grid()

        # Source and Destination images
        label_dst = Tkinter.Label(self)
        label_dst.grid(row=0, column=0)
        self.dst_img_manager.set_tk_label(label_dst)

        Tkinter.Label(self, text="+").grid(row=0, column=1)

        label_src = Tkinter.Label(self)
        label_src.grid(row=0, column=2)
        self.src_img_manager.set_tk_label(label_src)

        # Destination image buttons (Move, Rotate)
        dst_img_buttons = Tkinter.Frame(self)
        dst_img_buttons.grid(row=1, column=0)
        dst_edit_mode = Tkinter.StringVar()
        dst_edit_mode.trace('w', self.dst_img_manager.mode_changed)
        dst_edit_mode.set('move')
        for text, mode in self.dst_img_manager.EDIT_MODES:
            b = Tkinter.Radiobutton(
                dst_img_buttons, text=text, variable=dst_edit_mode, value=mode,
                indicatoron=0)
            b.pack(side=Tkinter.LEFT)
        self.dst_img_manager.set_edit_mode_str(dst_edit_mode)

        # Source image buttons (Draw/Erase/Move)
        src_img_buttons = Tkinter.Frame(self)
        src_img_buttons.grid(row=1, column=2)
        src_edit_mode = Tkinter.StringVar()
        src_edit_mode.set('draw')

        for text, mode in self.src_img_manager.EDIT_MODES:
            b = Tkinter.Radiobutton(src_img_buttons, text=text,
                                    variable=src_edit_mode, value=mode,
                                    indicatoron=0)
            b.pack(side=Tkinter.LEFT)
        self.src_img_manager.set_edit_mode_str(src_edit_mode)

        b = Tkinter.Button(src_img_buttons, text='Clear',
                           command=self.src_img_manager.clear_mask)
        b.pack(side=Tkinter.LEFT)

        # Size buttons
        def _draw_size_buttons(row, column, functions):
            size_buttons = Tkinter.Frame(self)
            size_buttons.grid(row=row, column=column)
            plus_button = Tkinter.Button(size_buttons, text='+', width=2,
                                         command=functions['+'])
            original_buttton = Tkinter.Button(size_buttons, text='100%',
                                              width=5,
                                              command=functions['original'])
            minus_button = Tkinter.Button(size_buttons, text='-', width=2,
                                          command=functions['-'])
            minus_button.pack(side=Tkinter.LEFT)
            original_buttton.pack(side=Tkinter.LEFT)
            plus_button.pack(side=Tkinter.LEFT)

        # Size buttons for source image
        _draw_size_buttons(2, 2, self.src_img_manager.ZOOM_FUNCTIONS)

        # Blend button
        action_frame = Tkinter.Frame(self)
        action_frame.grid(row=3, column=0, columnspan=3)

        if SAVE_MASK_ENABLED:
            def save_mask():
                self.src_img_manager.save_mask_image()
                print self.dst_img_manager.offset

            Tkinter.Button(action_frame, text='Save mask image',
                           command=save_mask).pack()

        Tkinter.Button(action_frame, text=u'Blend', command=self.blend).pack()

    def create_menu(self):
        menu_bar = Tkinter.Menu(self)

        file_menu = Tkinter.Menu(menu_bar)
        file_menu.add_command(label='Open Source Image',
                              command=self.src_img_manager.open_from_dialog)
        file_menu.add_command(label='Open Destination Image',
                              command=self.dst_img_manager.open_from_dialog)
        menu_bar.add_cascade(label="File", menu=file_menu)

        run_menu = Tkinter.Menu(menu_bar)
        run_menu.add_command(label='Blend!', command=self.blend)
        menu_bar.add_cascade(label="Run", menu=run_menu)

        example_menu = Tkinter.Menu(menu_bar)
        for i in range(self.NUM_OF_EXAMPLES):
            i += 1
            example_menu.add_command(label="Example %d" % i,
                                     command=self.load_example(i))
        menu_bar.add_cascade(label="Examples", menu=example_menu)

        self.config(menu=menu_bar)

    def load_example(self, example_id):
        """
         Load an example to GUI.
        :param example_id: id starts with *one*
        """

        def _load_example():
            src_path = rp('./testimages/test%d_src.png' % example_id)
            dst_path = rp('./testimages/test%d_target.png' % example_id)
            mask_path = None

            try:
                from pypoi.testimages.config import offset

                self.dst_img_manager.offset = offset[example_id - 1]
                self.dst_img_manager.rotate = 0
                mask_path = rp('./testimages/test%d_mask.png' % example_id)
            except IndexError:
                pass

            self.src_img_manager.open(src_path, mask_path=mask_path)
            self.dst_img_manager.open(dst_path)

        return _load_example

    def blend(self):
        angle = self.dst_img_manager.rotate
        src = np.asarray(self.src_img_manager.image_src.rotate(angle))
        src.flags.writeable = True
        dst = np.asarray(self.dst_img_manager.image)
        dst.flags.writeable = True
        mask = np.asarray(self.src_img_manager.image_mask.rotate(angle))
        mask.flags.writeable = True

        # poissonblending.blend takes (y, x) as offset,
        # whereas gui has (x, y) as offset values so reverse these values.
        reversed_offset = self.dst_img_manager.offset[::-1]
        blended_image = poissonblending.blend(dst, src, mask, reversed_offset)
        self.image_result = PIL.Image.fromarray(np.uint8(blended_image))
        self.image_tk_result = PIL.ImageTk.PhotoImage(self.image_result)

        result_window = Tkinter.Toplevel()
        label = Tkinter.Label(result_window, image=self.image_tk_result)
        label.image = self.image_tk_result  # for holding reference counter
        label.pack()
        save_button = Tkinter.Button(result_window, text='Save',
                                     command=self.save_result(result_window))
        save_button.pack()

        result_window.title("Blended Result")

    def save_result(self, parent):
        def _save_result():
            file_name = tkFileDialog.asksaveasfilename(parent=parent)
            try:
                self.image_result.save(file_name)
            except KeyError:
                msg = 'Unknown extension. Supported extensions:\n'
                msg += ' '.join(PIL.Image.EXTENSION.keys())
                tkMessageBox.showerror("Error", msg)

        return _save_result


def main():
    global app
    logging.basicConfig(level=logging.INFO)
    app = PoissonBlendingApp(None)
    app.title('PyPoi: "Py"thon Program for "Poi"sson Image Editing')
    app.mainloop()


if __name__ == "__main__":
    main()



import os
from setuptools import setup

desciption = '"Py"thon program for "Poi"sson Image Editing'
if os.path.exists('README.txt'):
    long_description = open('README.txt').read()
else:
    long_description = desciption

setup(
    name='pypoi',
    version='0.2.0',
    description=desciption,
    long_description=long_description,
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
    ],

    author='Fumihiro Bessho (Ben)',
    author_email='fumihiro.bessho@gmail.com',
    url='https://github.com/fbessho/PyPoi',
    packages=['pypoi', 'pypoi.testimages'],
    package_data={'pypoi': ['testimages/*.png']},
    scripts=['scripts/pypoi']
)



import os
import sys


def resource_path(relative):
    """
    Resolve path to resource (for supporting PyInstaller).

    When PyInstaller is executed with --onefile option, resources are put in a
    temporary dir and scripts need to use the info to access resources.

    ref: http://stackoverflow.com/questions/13946650/pyinstaller-2-0-bundle-file-as-onefile
    """
    if hasattr(sys, '_MEIPASS'):
        root = os.path.join(sys._MEIPASS, 'pypoi')
    else:
        root = os.path.abspath(os.path.dirname(__file__))

    return os.path.join(root, relative)



import logging
import tkMessageBox
import PIL.Image
import PIL.ImageTk
import PIL.ImageChops
import PIL.ImageDraw
import tkFileDialog
import math
from collections import namedtuple


EditMode = namedtuple('EditMode', 'display value')


class ImageManager():
    def __init__(self):
        pass

    def open_from_dialog(self):
        path = tkFileDialog.askopenfilename()
        if len(path) > 0:
            self.open(path)

    def open(self, path, **kwargs):
        self.path = path
        self.load(**kwargs)
        self.draw()

    def set_path(self, path):
        self.path = path

    def draw(self):
        """Called when GUI needs to be refreshed"""

    def on_mouse_move2(self):
        pass

    def set_tk_label(self, tk_label):
        tk_label.bind('<Button-1>', self.on_mouse_down)
        tk_label.bind('<B1-Motion>', self.on_mouse_move)
        tk_label.bind('<B2-Motion>', self.on_mouse_move2)
        self.tk_label = tk_label


class SourceImageManager(ImageManager):
    """Manage source image and mask image, which is shown on the right hand
    side in the GUI.

    Attributes:
        self.path: path to source image file.
        self.image_src: PIL.Image object of source image.
        self.image_mask: PIL.Image object of mask image.
        self.edit_mode: Tkinter.StringVar(). Stores current mode selected.
    """
    EDIT_MODES = (
        EditMode('Draw', 'draw'),
        EditMode('Erase', 'erase'),
        EditMode('Move', 'move'),
    )

    def __init__(self):
        ImageManager.__init__(self)
        self.draw_propagate_functions = []
        self.ZOOM_FUNCTIONS = {
            '+': self.zoom_by_ten_percent,
            '-': self.shrink_by_ten_percent,
            'original': self.reset_size
        }

        self.path = None
        self.image_src = None
        self.image_mask = None
        self.edit_mode = None

    def set_edit_mode_str(self, edit_mode_str):
        self.edit_mode = edit_mode_str

    def add_propagation_func(self, callback_func):
        self.draw_propagate_functions.append(callback_func)

    def load(self, mask_path=None):
        self.image_src = PIL.Image.open(self.path).convert("RGB")
        if mask_path:
            self.image_mask = PIL.Image.open(mask_path).convert("L")
        else:
            self.image_mask = PIL.Image.new('L', self.image_src.size)

        self.original_size = self.image_src.size
        self.current_scale_percentage = 100

    def draw(self):
        self.image_masked_src = PIL.Image.blend(self.image_src,
                                                self.image_mask.convert("RGB"),
                                                0.3)
        self.image_tk_masked_src = PIL.ImageTk.PhotoImage(self.image_masked_src)
        self.tk_label.configure(image=self.image_tk_masked_src)
        for f in self.draw_propagate_functions:
            f()

    def on_mouse_down(self, event):
        if self.edit_mode.get() == 'move':
            print "Mouse button down:", event.x, event.y
            self.sx, self.sy = event.x, event.y
        elif self.edit_mode.get() == 'draw':
            self.modify_mask(event.x, event.y, 255)
        elif self.edit_mode.get() == 'erase':
            self.modify_mask(event.x, event.y, 0)

    def on_mouse_move(self, event):
        if self.edit_mode.get() == 'move':
            print "Mouse move:", event.x, event.y
            dx = event.x - self.sx
            dy = event.y - self.sy
            self.image_mask = PIL.ImageChops.offset(self.image_mask, dx, dy)
            self.draw()

            self.sx, self.sy = event.x, event.y
        elif self.edit_mode.get() == 'draw':
            self.modify_mask(event.x, event.y, 255)
        elif self.edit_mode.get() == 'erase':
            self.modify_mask(event.x, event.y, 0)

    def modify_mask(self, x, y, new_value):
        pixel = self.image_mask.load()
        mx, my = self.image_mask.size
        for dx in range(-10, 11):
            for dy in range(-10, 11):
                nx = x + dx
                ny = y + dy
                if 0 <= nx < mx and 0 <= ny < my:
                    pixel[x + dx, y + dy] = new_value
        self.draw()

    def resize(self, new_scale):
        self.image_src = PIL.Image.open(self.path).convert("RGB")
        new_size = tuple([int(x * new_scale) for x in self.image_src.size])
        self.image_src = self.image_src.resize(new_size)
        self.image_mask = self.image_mask.resize(new_size)
        self.draw()

    def shrink_by_ten_percent(self):
        self.current_scale_percentage -= 10
        self.resize(self.current_scale_percentage / 100.0)

    def zoom_by_ten_percent(self):
        self.current_scale_percentage += 10
        self.resize(self.current_scale_percentage / 100.0)

    def reset_size(self):
        self.current_scale_percentage = 100
        self.resize(1)

    def clear_mask(self):
        yn = tkMessageBox.askokcancel('Clear mask image',
                                      'Mask will be cleared, do you proceed?')
        if yn:
            self.image_mask = PIL.Image.new('L', self.image_src.size)
            self.draw()

    def save_mask_image(self):
        self.image_mask.save('mask.png')


class DestinationImageManager(ImageManager):
    """Manage destination image, the one on the left hand side in the GUI.

    Attributes:
        rotate: In degrees
        edit_mode: Tkinter.StringVar(). Stores current mode selected.
        image: A destination image as a PIL.Image object.
        src_img_cropped: Source image cropped to mask area.
        src_mask_cropped: Mask image cropped to mask area.
    """
    logger = logging.getLogger('DestinationImageManager')
    EDIT_MODE_MOVE = EditMode('Move', 'move')
    EDIT_MODE_ROTATE = EditMode('Rotate', 'rotate')
    EDIT_MODES = (EDIT_MODE_MOVE, EDIT_MODE_ROTATE)

    def __init__(self, src_img_manager):
        ImageManager.__init__(self)
        self.src_img_manager = src_img_manager
        self.src_img_manager.add_propagation_func(self.draw)

        self.image = None
        self.src_img_cropped = None
        self.src_mask_cropped = None
        self.offset = (0, 0)
        self.rotate = 0
        self.edit_mode = None

    def load(self):
        self.image = PIL.Image.open(self.path).convert("RGB")

    def draw(self):
        image_to_show = self.image.copy()
        self.draw_mask(image_to_show)

        self.image_tk = PIL.ImageTk.PhotoImage(image_to_show)
        self.tk_label.configure(image=self.image_tk)

    def draw_mask(self, image):
        """Add mask to image"""

        img_src = self.src_img_manager.image_src.copy()
        mask = self.src_img_manager.image_mask.copy()

        # Draw box around mask
        mask_bbox = self.src_img_manager.image_mask.getbbox()
        l, u, r, b = mask_bbox
        mask_bbox = l+1, u+1, r-1, b-1  # shrink bbox by 1px to avoid run out of edge
        if (mask_bbox is not None and
            self.edit_mode.get() == self.EDIT_MODE_ROTATE.value):
            draw = PIL.ImageDraw.Draw(img_src)
            draw.rectangle(mask_bbox, None, 'red')

            draw = PIL.ImageDraw.Draw(mask)
            draw.rectangle(mask_bbox, None, 'white')

        squared_msk_img = SquareMaskImage(img_src, mask, self.offset, self.rotate)

        image.paste(
            squared_msk_img.squared_src_img,
            squared_msk_img.offset,
            squared_msk_img.squared_msk_img
        )

    def set_edit_mode_str(self, edit_mode):
        self.edit_mode = edit_mode

    def on_mouse_down(self, event):
        self.sx, self.sy = event.x, event.y

    def on_mouse_move(self, event):

        if self.edit_mode.get() == self.EDIT_MODE_MOVE.value:
            dx, dy = (event.x - self.sx, event.y - self.sy)
            x, y = self.offset
            self.offset = (x + dx, y + dy)

        elif self.edit_mode.get() == self.EDIT_MODE_ROTATE.value:
            x0, y0 = self.calc_center_of_mask()
            rotate_old = self.calc_angle(x0, y0, self.sx, self.sy)
            rotate_new = self.calc_angle(x0, y0, event.x, event.y)
            self.rotate += rotate_new - rotate_old

        else:
            self.logger.error('Invalid edit mode: %s' % self.edit_mode.get())

        self.draw()
        self.sx, self.sy = event.x, event.y

    def calc_angle(self, x0, y0, x1, y1):
        """Calc angle from (x0, y0) to (x1, y1).

        x0, y0, x1, y1 is provided as PIL coordinates (top-left is the root),
        however return value is angle in the normal mathematical coordinates
        (bottom-left is the root).

        Returns:
            Angle in degrees
        """
        y0, y1 = -y0, -y1
        angle_in_radian = math.atan2((y1-y0), (x1-x0))
        return angle_in_radian / math.pi * 180.0

    def calc_center_of_mask(self):
        """Returns center of the mask image."""
        left, upper, right, lower = self.src_img_manager.image_mask.getbbox()
        return (left+right)/2.0, (upper+lower)/2.0

    def mode_changed(self, *args, **kwargs):
        """Callback function which is called when mode changes"""
        if self.image:
            self.draw()


class SquareMaskImage():
    def __init__(self, src_img, msk_img, offset, rotate):
        self.squared_src_img, self.squared_msk_img = self.create_images(src_img, msk_img)
        self.squared_src_img = self.squared_src_img.rotate(rotate)
        self.squared_msk_img = self.squared_msk_img.rotate(rotate)

        self.offset = self.calc_new_offset(msk_img, offset)

    def create_images(self, src_img, mask_img):
        left, upper, right, lower = mask_img.getbbox()
        square_size = int(math.sqrt((right-left)**2 + (upper-lower)**2))  # create square of this px on a side
        cropped_mask_image = mask_img.crop(mask_img.getbbox())
        cropped_src_image = src_img.crop(mask_img.getbbox())

        square_src_image = PIL.Image.new(src_img.mode, (square_size, square_size))
        square_mask_image = PIL.Image.new(mask_img.mode, (square_size, square_size))

        center_of_square = (square_size/2, square_size/2)
        top_left = (center_of_square[0] - (right-left)/2), (center_of_square[1] - (lower-upper)/2)
        square_src_image.paste(cropped_src_image, top_left)
        square_mask_image.paste(cropped_mask_image, top_left)

        return square_src_image, square_mask_image

    def calc_new_offset(self, msk_img, original_offset):
        left, upper, right, lower = msk_img.getbbox()
        center_of_bbox = (left+right)/2, (upper+lower)/2
        square_size = int(math.sqrt((right-left)**2 + (upper-lower)**2))
        offset_delta = center_of_bbox[0] - square_size/2, center_of_bbox[1] - square_size/2

        return (original_offset[0] + offset_delta[0]), (original_offset[1] + offset_delta[1])
